package gov.va.nvap.server.endpoint.vista.link;

import gov.va.med.exception.FoundationsException;
import gov.va.med.vistalink.adapter.cci.VistaLinkConnection;
import gov.va.med.vistalink.rpc.RpcRequest;
import gov.va.med.vistalink.rpc.RpcRequestFactory;
import gov.va.med.vistalink.rpc.RpcRequestParams;
import gov.va.med.vistalink.rpc.RpcResponse;
import gov.va.med.vistalink.security.VistaKernelPrincipalImpl;
import gov.va.nvap.common.auth.context.LoginContextExt;
import gov.va.nvap.common.endpoint.EndpointException;
import gov.va.nvap.common.validation.NullChecker;

import java.util.List;
import java.util.Map;

import javax.resource.ResourceException;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.Configuration;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.springframework.beans.factory.annotation.Required;

/**
 * VistALink Endpoint to make RPC calls into VistA systems.
 * 
 * @author Asha Amritraj
 */
public class VistALinkEndpoint implements
		gov.va.nvap.common.endpoint.Endpoint<Object, String> {

	private CallbackHandler callbackHandler;
	private Configuration configuration;
	private String loginContextName;
	private Map<Integer, String> parameterTypes;
	private String rpcContext;
	private String rpcName;
	private int timeout;

	public String getRpcContext() {
		return this.rpcContext;
	}

	public String getRpcName() {
		return this.rpcName;
	}

	public int getTimeout() {
		return this.timeout;
	}

	@Override
	public String invoke(final Object args) throws EndpointException {

		VistaLinkConnection connection = null;
		LoginContext loginContext = null;
		try {
			// Login using the access/verify code set in the callback handler
			loginContext = new LoginContextExt(this.loginContextName,
					this.callbackHandler, this.configuration);
			loginContext.login();
			final VistaKernelPrincipalImpl userPrincipal = VistaKernelPrincipalImpl
					.getKernelPrincipal(loginContext.getSubject());
			connection = userPrincipal.getAuthenticatedConnection();
			connection.setTimeOut(this.timeout);
			final RpcRequest rpcRequest = RpcRequestFactory.getRpcRequest(
					this.rpcContext, this.rpcName);
			final RpcRequestParams params = rpcRequest.getParams();

			if (List.class.isInstance(args)) {
				final List<?> argList = (List<?>) args;
				for (int i = 0; i < argList.size(); i++) {
					params.setParam(i + 1, this.parameterTypes.get(i + 1),
							argList.get(i));
				}
			} else if (String.class.isInstance(args)) {
				final String argValue = (String) args;
				params.setParam(1, this.parameterTypes.get(1), argValue);
			}

			final RpcResponse rpcResponse = connection.executeRPC(rpcRequest);
			final String results = rpcResponse.getResults();
			return results;
		} catch (final LoginException ex) {
			throw new RuntimeException(ex);
		} catch (final FoundationsException ex) {
			throw new RuntimeException(ex);
		} finally {
			if (NullChecker.isNotEmpty(loginContext)) {
				try {
					// Always logout
					loginContext.logout();
				} catch (final LoginException ex) {
					// TODO: Log!
					// Ignore Exception!!
				}
			}
			if (NullChecker.isNotEmpty(connection)) {
				try {
					connection.close();
				} catch (final ResourceException ex) {
					// TODO: Log!
					// Ignore Exception!!
				}
			}
		}
	}

	@Required
	public void setCallbackHandler(final CallbackHandler callbackHandler) {
		this.callbackHandler = callbackHandler;
	}

	@Required
	public void setConfiguration(final Configuration configuration) {
		this.configuration = configuration;
	}

	@Required
	public void setLoginContextName(final String loginContextName) {
		this.loginContextName = loginContextName;
	}

	@Required
	public void setParameterTypes(final Map<Integer, String> parameterTypes) {
		this.parameterTypes = parameterTypes;
	}

	@Required
	public void setRpcContext(final String rpcContext) {
		this.rpcContext = rpcContext;
	}

	@Required
	public void setRpcName(final String rpcName) {
		this.rpcName = rpcName;
	}

	@Required
	public void setTimeout(final int timeout) {
		this.timeout = timeout;
	}
}
