package gov.va.nvap.server.service.privacy.pdf;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.apache.pdfbox.exceptions.COSVisitorException;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.edit.PDPageContentStream;
import org.apache.pdfbox.pdmodel.font.PDFont;
import org.apache.pdfbox.pdmodel.font.PDType1Font;


/**
 * @author Sridhar Varma Alluri
 * @author Zack Peterson
 * This class is a helper class to generate a pdf by replacing a template string with actual values 
 */
public class PdfGenerator {
	private PDDocument document = null;
	private PDFont font = null;
	private int begin_x = 0;
	private PDRectangle rect = null;
	private int line = 0;
	private int lineSpacing = 15;
	private float begin_y = 0;
    
    private final int PAGE_WIDTH = 105;
    private final int INDENT_WIDTH = 65;
    private final int MAX_LINES = 45;

	/**
	 * 
	 * Default constructor initializing member variables  
	 */
	public PdfGenerator() {
		document = new PDDocument();
		font = null;
		begin_x = 0;
		rect = null;
		line = 0;
		lineSpacing = 15;
		begin_y = 0;
	}

	/**
	 * Creates a pdf document from template and patientsDetails. 
	 * A page is created for each patient in the collection.
	 * @param  template  The template that is used to create the pdf document
	 * @param  patientsDetails a collection of patient details that are replaced in template
	 * @return a stream that has the pdf document
	 */
	public ByteArrayOutputStream create(String template,
			ArrayList<HashMap<String, String>> patientsDetails) {
		ByteArrayOutputStream ostream = new ByteArrayOutputStream();
		// Split up the template into individual lines 
		String[] lines = template.split("\n");
        // Split lines at space nearest page width limit
        ArrayList<String> adjustedLines = new ArrayList<String>();
        for (String str : lines) {
            while(str.length() > PAGE_WIDTH) {
                int spaceDelim = str.lastIndexOf(32, PAGE_WIDTH);
                adjustedLines.add(str.substring(0, spaceDelim));
                str = str.substring(spaceDelim + 1);
            }
            adjustedLines.add(str);
        }
        
		try {
			for (HashMap<String, String> hashMap : patientsDetails) {
				PDPage page = addPage(adjustedLines, hashMap);
				document.addPage(page);
			}
			document.save(ostream);
			document.close();
		} catch (COSVisitorException e) {
			ostream = null;
		} catch (IOException e) {
			ostream = null;
		}
		return ostream;
	}

    /**
     * Creates a pdf document from multiple templates and patientsDetails collections.
     * A page is created for each patient in the collections.
     *
     * @param templateArray The templates that are used to create the pdf document
     * @param patientsDetailsList A List of patient details that are replaced in the templates
     * @return a stream that has the pdf document
     */
    public ByteArrayOutputStream create(String[] templateArray,
        ArrayList<ArrayList<HashMap<String, String>>> patientsDetailsList) {
        ByteArrayOutputStream ostream = new ByteArrayOutputStream();
        
        // there must be exactly one template for each patient detail collection
        if (templateArray.length != patientsDetailsList.size()) {
            return null;
        }
        
        for (int i = 0; i < templateArray.length; i++) {
            String template = templateArray[i];
            ArrayList<HashMap<String, String>> patientsDetails = patientsDetailsList.get(i);
            
            // Split up the template into individual lines 
            String[] lines = template.split("\n");
            // Split lines at space nearest page width limit
            ArrayList<String> adjustedLines = new ArrayList<String>();
            for (String str : lines) {
                while (str.length() > PAGE_WIDTH) {
                    int spaceDelim = str.lastIndexOf(32, PAGE_WIDTH);
                    adjustedLines.add(str.substring(0, spaceDelim));
                    str = str.substring(spaceDelim + 1);
                }
                adjustedLines.add(str);
            }

            try {
                for (HashMap<String, String> hashMap : patientsDetails) {
                    PDPage page = addPage(adjustedLines, hashMap);
                    document.addPage(page);
                }
            } catch (IOException e) {
                ostream = null;
            }
        }
        
        try {
            document.save(ostream);
            document.close();
        } catch (COSVisitorException e) {
            ostream = null;
        } catch (IOException e) {
            ostream = null;
        }
        
        return ostream;
    }
        
	/**
	 * A pdf page per patient is created from template. 
	 * A page is created for each patient in the collection.
	 * @param  lines  An array of each individual line
	 * @param  patientDetail hashmap with patient information
	 * @return a stream that has the pdf document
	 */
	private PDPage addPage(ArrayList<String> lines, HashMap<String, String> patientDetail)
			throws IOException {
		PDPage page = new PDPage();
		font = PDType1Font.TIMES_ROMAN;

		PDPageContentStream contentStream = new PDPageContentStream(document,
				page);
		line = 0;
		begin_x = INDENT_WIDTH;
		rect = page.getMediaBox();
		begin_y = rect.getHeight() - INDENT_WIDTH;

        int i = 0;
		for (String str : lines) {
            i++; if(i > MAX_LINES) break;
			contentStream.beginText();
			contentStream.setFont(font, 11);
			contentStream.moveTextPositionByAmount(begin_x, begin_y
					- lineSpacing * (++line));
			contentStream.drawString(getUpdatedString(str, patientDetail));
			contentStream.endText();
		}
		contentStream.close();
		return page;
	}

	/**
	 * This method replaces string that has the keys with  patient information. 
	 * @param  templateString  Template string that has keys
	 * @param  patientDetail a collection of patient details that are replaced in template
	 * @return String that has keys replaced with values
	 */
	private String getUpdatedString(String templateString,
			HashMap<String, String> patientDetail) {
		String updatedString = templateString;
		for (Map.Entry<String, String> entry : patientDetail.entrySet()) {
			String key = entry.getKey();
			String value = entry.getValue();
			updatedString = updatedString.replace(key, value);
		}
		return updatedString;
	}

}
