package gov.va.nvap.svc.consentmgmt.stub.data;

import gov.va.nvap.common.sql.SqlUtil;
import gov.va.nvap.common.validation.NullChecker;
import java.io.Serializable;
import java.util.Date;

import javax.persistence.Basic;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.NamedNativeQueries;
import javax.persistence.NamedNativeQuery;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

/**
 *
 * @author 
 * @author 
 * edited by                  
 */
@Entity
@Table(name = "PATIENT_CONSENT_DIR")
@NamedNativeQueries({
    @NamedNativeQuery(
            name="DetailedConsentDirective.findAllExpiringConsentDirectives",
            query=DetailedConsentDirective.findAllExpiringConsentDirectivesNativeSQL,
            resultClass=DetailedConsentDirective.class
    )
})

public class DetailedConsentDirective implements Serializable {
	private static final long serialVersionUID = 1L;
    
	/**
	 * String constant native SQL for retrieving patient consent directives that are set to
	 * expire in a given number of days whose patients have NOT received a notification
	 */
	protected static final String findAllExpiringConsentDirectivesNativeSQL=
    "select * from ( " +
    "    select A.CONSENT_DIR_ID, " +
    "        A.OPTIN_CONSENT_TYPE_ID, " +
    "        A.OPTIN_DATE, " +
    "        A.EXPIRATION_DATE, " +
    "        A.PATIENT_IEN, " +
    "        J.CONSENT_TYPE_ID_FROM_AUDIT, " +
    "        J.PATIENT_SSN, " +
    "        J.PATIENT_LAST_NAME, " +
    "        J.PATIENT_GIVEN_NAME, " +
    "        J.PATIENT_MIDDLE_NAME, " +
    "        J.FACILITY_NAME, " +
    "        J.USER_ID, " +
    "        ROW_NUMBER () over (order by CONSENT_DIR_ID ASC) AS LINE_NUMBER, " + // Hardcoded default sort to keep SQL valid since we cannot bind column names to variables
    "        count(1) over () TOTAL_ROWS " +
    "    from PATIENT_CONSENT_DIR A inner join ( " +
    "        select * " +
    "        from ( " +
    "            select PATIENT_ID, " +
    "                PATIENT_SSN, " +
    "                PATIENT_LAST_NAME, " +
    "                PATIENT_GIVEN_NAME, " +
    "                PATIENT_MIDDLE_NAME, " +
    "                FACILITY_NAME, " +
    "                USER_ID, " +
    "                ROW_NUMBER () over (partition by PATIENT_ID, CONSENT_TYPE order by CREATED_DATE desc) as RowNumber, " +
    "                CT.CONSENT_TYPE_ID as CONSENT_TYPE_ID_FROM_AUDIT " +
    "            from CONSENT_AUDIT inner join CONSENT_TYPE CT on CT.NAME = CONSENT_TYPE " +
    "                where ACTION_TYPE = 'OPT-IN' " +
    "                    and (CONSENT_TYPE = 'NwHIN Authorization' or CONSENT_TYPE = 'SSA Authorization') " +
    "                    /*patient_type_filter*/ " +
    "        ) " +
    "        where RowNumber = 1 " +
                 // Filtering by columns that aren't in "partition by" clause needs to take place outside of partitioning 
                 // since we always take the first row in each partition
    "            /*user_id_filter*/ " + // 
    "            and (:facility_name is null or FACILITY_NAME = :facility_name)" +
    "    ) J on A.PATIENT_IEN = J.PATIENT_ID " +
    "    where OPTIN_CONSENT_TYPE_ID is not null " +
    "        and OPTOUT_DATE is null " +
    "        and OPTOUT_TS is null " +
    "        and OPTIN_CONSENT_TYPE_ID = CONSENT_TYPE_ID_FROM_AUDIT " + // This clause will filter out any inconsistencies between PATIENT_CONSENT_DIR and CONSENT_AUDIT tables
    "        and (:consent_type_id is null or OPTIN_CONSENT_TYPE_ID = :consent_type_id) " +
    "        and (:end_date is null OR EXPIRATION_DATE <= to_date(:end_date, 'YYYY-MM-DD')) " +
    "        and ((:start_date is null AND EXPIRATION_DATE > systimestamp) OR EXPIRATION_DATE >= to_date(:start_date, 'YYYY-MM-DD')) " +
    ") " +
    "where (LINE_NUMBER between :from_page and :to_page) or (:to_page < 0)";

	@Id
	@GeneratedValue(strategy = GenerationType.AUTO)
	@Basic(optional = false)
	@Column(name = "CONSENT_DIR_ID")
	private Long consentDirId;
	@ManyToOne(fetch = FetchType.EAGER)
	@JoinColumn(name = "OPTIN_CONSENT_TYPE_ID")
	private ConsentType optinConsentType;
	@Column(name = "EXPIRATION_DATE")
	@Temporal(TemporalType.TIMESTAMP)
	private Date expirationDate;
	@Basic(optional = false)
	@Column(name = "OPTIN_DATE")
	@Temporal(TemporalType.TIMESTAMP)
	private Date optinDate;
    @Column(name = "TOTAL_ROWS")
	private Long totalRows;
    
    // Info from CONSENT_AUDIT
    @Column(name = "PATIENT_SSN")
    private String ssn;
    @Column(name = "PATIENT_LAST_NAME")
    private String lastName;
    @Column(name = "PATIENT_GIVEN_NAME")
    private String firstName;
    @Column(name = "PATIENT_MIDDLE_NAME")
    private String middleName;
    @Column(name = "FACILITY_NAME")
    private String facilityName;
    @Column(name = "USER_ID")
    private String userId;
    
    public static String constructQuery(String userId, int patientTypes, String sortColumn, String sortDirection) {
        String query = findAllExpiringConsentDirectivesNativeSQL;
        
        // We cannot bind variables to column names, so we have a hardcoded default sorting clause to keep native query valid.
        // Here, we replace the sorting clause with the values received from the UI.
        if (sortColumn.indexOf("_DATE") > 0 || sortColumn.indexOf("_TS") > 0) {
            query = query.replace("(order by CONSENT_DIR_ID ASC)", "(order by " + sortColumn + " " + sortDirection + ")");
        } else {
            query = query.replace("(order by CONSENT_DIR_ID ASC)", "(order by nlssort(" + sortColumn + ", 'NLS_SORT = WEST_EUROPEAN') " + sortDirection + ")");
        }
        
        // Filter by "Entered By", i.e. user ID
        String userIdFilter;
        if (NullChecker.isEmpty(userId)) {
            userIdFilter = " and :user_id is null ";
        } else if (userId.equalsIgnoreCase("eBenefits")) { // or this with "MHV" when MHV becomes its own filter value
            userIdFilter = " and USER_ID = :user_id ";
        } else {
            // Entered by an employee
            userIdFilter = " and (:user_id is null or USER_ID <> 'eBenefits') ";
        }
        query = query.replace("/*user_id_filter*/", userIdFilter);
        
        // Filter by patient type (real=1, test=2, both=3)
        if (patientTypes == 1) {
            // Exclude test patients
            query = query.replace("/*patient_type_filter*/", " and " + SqlUtil.generateExcludeTestPatientsClause(false, "CONSENT_AUDIT", "PATIENT_SSN"));
        }
        
        return query;
    }
    
    public String getLastName() {
        return lastName;
    }

    public void setLastName(String lastName) {
        this.lastName = lastName;
    }

    public String getFirstName() {
        return firstName;
    }

    public void setFirstName(String firstName) {
        this.firstName = firstName;
    }

    public String getMiddleName() {
        return middleName;
    }

    public void setMiddleName(String middleName) {
        this.middleName = middleName;
    }

    public String getSsn() {
        return ssn;
    }

    public void setSsn(String ssn) {
        this.ssn = ssn;
    }

    public String getFacilityName() {
        return facilityName;
    }

    public void setFacilityName(String facilityName) {
        this.facilityName = facilityName;
    }

    public String getUserId() {
        return userId;
    }

    public void setUserId(String userId) {
        this.userId = userId;
    }

	public DetailedConsentDirective() {
	}

	public DetailedConsentDirective(final Long consentDirId) {
		this.consentDirId = consentDirId;
	}

	public DetailedConsentDirective(final Long consentDirId, final String patientIen,
			final Date optinDate, final Date expirationDate,
			final Date optinTS, final ConsentType consentType) {
		this.consentDirId = consentDirId;
		this.optinDate = optinDate;
		this.expirationDate = expirationDate;
		this.optinConsentType = consentType;
	}

	@Override
	public boolean equals(final Object object) {
		// TODO: Warning - this method won't work in the case the id fields are
		// not set
		if (!(object instanceof DetailedConsentDirective)) {
			return false;
		}
		final DetailedConsentDirective other = (DetailedConsentDirective) object;
		if (((this.consentDirId == null) && (other.consentDirId != null))
				|| ((this.consentDirId != null) && !this.consentDirId
						.equals(other.consentDirId))) {
			return false;
		}
		return true;
	}

	public Long getConsentDirId() {
		return this.consentDirId;
	}

	public Date getExpirationDate() {
		return this.expirationDate;
	}

	public Date getOptinDate() {
		return this.optinDate;
	}

	@Override
	public int hashCode() {
		int hash = 0;
		hash += (this.consentDirId != null ? this.consentDirId.hashCode() : 0);
		return hash;
	}

	public void setConsentDirId(final Long consentDirId) {
		this.consentDirId = consentDirId;
	}

	public void setExpirationDate(final Date expirationDate) {
		this.expirationDate = expirationDate;
	}

	public void setOptinDate(final Date optinDate) {
		this.optinDate = optinDate;
	}

	@Override
	public String toString() {
		return "gov.va.med.vler.nhin.vap.pip.entities.DetailedConsentDirective[consentDirId="
				+ this.consentDirId + "]";
	}

	public void setOptinConsentType(ConsentType optinConsentType) {
		this.optinConsentType = optinConsentType;
	}

	public ConsentType getOptinConsentType() {
		return optinConsentType;
	}

	public Long getTotalRows() {
		return totalRows;
	}

	public void setTotalRows(Long totalRows) {
		this.totalRows = totalRows;
	}
	
}
