package gov.va.nvap.web.consent.audit.dao;

import gov.va.nvap.common.sql.SqlUtil;
import gov.va.nvap.common.validation.Assert;
import gov.va.nvap.common.validation.NullChecker;
import gov.va.nvap.web.consent.audit.AuditedConsentSummary;

import java.io.IOException;
import java.io.StringReader;
import java.security.InvalidParameterException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;
import javax.persistence.TemporalType;

import org.hibernate.CacheMode;
import org.hibernate.FlushMode;

/**
 * DAO for getting the events from the CONSENT_AUDIT table.
 *
 * @author 
 * edited by               
 * edited by                  
 */
public class AuditedConsentDAO {

	/**
	 * Read buffer size.
	 */
	private static final int READ_BUFFER_SIZE = 4096;

	/**
	 * Entity Manager instance.
	 */
	@PersistenceContext
	private EntityManager em;

	private void appendGroupByClause(final StringBuffer buffer, final String str) {
		this.appendGroupByClause(buffer, str, " Group By ", " , ");
	}

	private void appendGroupByClause(final StringBuffer buffer,
			final String str, final String empty, final String notEmpty) {
		if (buffer.length() > 0) {
			buffer.append(notEmpty);
		} else {
			buffer.append(empty);
		}
		buffer.append(str);
	}

	private void appendWhereClause(final StringBuffer buffer, final String str) {
		this.appendWhereClause(buffer, str, " WHERE ", " AND ");
	}

	private void appendWhereClause(final StringBuffer buffer, final String str,
			final String empty, final String notEmpty) {
		if (buffer.length() > 0) {
			buffer.append(notEmpty);
		} else {
			buffer.append(empty);
		}
		buffer.append(str);
	}

	/**
	 * Get the events from the database table by calling ConsentAudit.findAll
	 * query.
	 */
        public final List<Object> getEvents(final Date startDate,
			final Date endDate, final int currentPage, final int pageSize) {

		final String reportQueryName = "ConsentAudit.findByDate";
		final Query query = this.em.createNamedQuery(reportQueryName)
				.setParameter("startDate", startDate, TemporalType.TIMESTAMP)
				.setParameter("endDate", endDate, TemporalType.TIMESTAMP);
		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		return this.getResults(updatedQuery, currentPage, pageSize);
	}

	/**
	 * Get the events from the audit log based on paramters.
	 */
        public final List<Object> getEvents(final Date startDate,
			final Date endDate, final String patientId, final int currentPage,
			final int pageSize) {

		final String reportQueryName = "ConsentAudit.findByPatientIdAndDate";
		final Query query = this.em.createNamedQuery(reportQueryName)
				.setParameter("startDate", startDate, TemporalType.TIMESTAMP)
				.setParameter("endDate", endDate, TemporalType.TIMESTAMP)
				.setParameter("patientId", patientId);
		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		return this.getResults(updatedQuery, currentPage, pageSize);
	}

	/**
	 * Get the events from the datbase table by building and calling a query.
	 * Use this method for complex queries of details. It accepts all parameters
	 * and builds the where clause based on values supplied.
	 */
        public final List<Object> getEvents(final Date startDate,
			final Date endDate, final String ssn, final String firstName,
			final String lastName, final String facilityName,
			final String visnName, final String consentType,
			final String inactivationReason, final String enteredBy,
			final int currentPage,
			final int pageSize, final String sortValue,
			final String sortDirection, final Boolean includeTestPatients) {

		final StringBuffer whereClause = new StringBuffer();
		final StringBuffer sortClause = new StringBuffer();
		final HashMap<String, Object> setParams = new HashMap<String, Object>();

		if (startDate != null) {
			this.appendWhereClause(whereClause, "a.timeOfEvent >= :startDate");
			setParams.put("startDate", startDate);
		}

		if (endDate != null) {
			this.appendWhereClause(whereClause, "a.timeOfEvent <= :endDate");
			setParams.put("endDate", endDate);
		}

		if (!NullChecker.isNullOrEmpty(ssn)) {
			this.appendWhereClause(whereClause, "a.patientSsn = :patientSsn");
			setParams.put("patientSsn", ssn);
		}

		// Task #411
		if (!NullChecker.isNullOrEmpty(firstName)) {
			this.appendWhereClause(whereClause,
					"upper(a.patientGivenName) = :firstName");
			setParams.put("firstName", firstName.toUpperCase());
		}

		// Task #411
		if (!NullChecker.isNullOrEmpty(lastName)) {
			this.appendWhereClause(whereClause,
					"upper(a.patientLastName) = :lastName");
			setParams.put("lastName", lastName.toUpperCase());
		}

		// task 300
		if (!NullChecker.isNullOrEmpty(facilityName)) {
			if (facilityName.equalsIgnoreCase("Unknown")) {
				this.appendWhereClause(whereClause, "a.facilityName IS NULL");
			} else {
				this.appendWhereClause(whereClause, "a.facilityName = :facilityName");
				setParams.put("facilityName", facilityName);
			}
		}

		if (!NullChecker.isNullOrEmpty(visnName)) {
			if (visnName.equalsIgnoreCase("Unknown")) {
				this.appendWhereClause(whereClause, "a.visnName IS NULL");
			} else {
				this.appendWhereClause(whereClause, "a.visnName = :visnName");
				setParams.put("visnName", visnName);
			}
		}

		if (!NullChecker.isNullOrEmpty(consentType)) {
			if (consentType.equalsIgnoreCase("allAuthorizations")){
				//or statement
				this.appendWhereClause(whereClause, "a.actionType = :actionType");
				setParams.put("actionType", "OPT-IN");
			} else if(consentType.equalsIgnoreCase("allRevocations")){
				//or statement
				this.appendWhereClause(whereClause, "a.actionType = :actionType");
				setParams.put("actionType", "OPT-OUT");
			} else if (consentType.equalsIgnoreCase("NwHINAuthorizations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Authorization");
			} else if (consentType.equalsIgnoreCase("SSAAuthorizations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "SSA Authorization");
			} else if (consentType.equalsIgnoreCase("NwHINRevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Revocation");
			} else if (consentType.equalsIgnoreCase("SSARevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "SSA Revocation");
			} else if (consentType.equalsIgnoreCase("NwHINRestrictions")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Organization Restriction Authorization");
			} else if (consentType.equalsIgnoreCase("NwHINRestrictionRevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Organization Restriction Revocation");
			} else {
				//TODO throw an error
				//program did not receive expected value
			}
		}
		
		if (!NullChecker.isNullOrEmpty(inactivationReason)) {
			this.appendWhereClause(whereClause,
					"a.optoutReason = :optoutReason");
			setParams.put("optoutReason", inactivationReason);
		}

		if (!NullChecker.isNullOrEmpty(enteredBy)){
			if(enteredBy.equalsIgnoreCase("eBenefits")){
				this.appendWhereClause(whereClause, "a.userId = :userId");
				setParams.put("userId", enteredBy);
			}else{ //it is employee
				//not eBenefits
				this.appendWhereClause(whereClause, "a.userId <> :userId");
				setParams.put("userId", "eBenefits");
			}
		}
        
        if (!Boolean.TRUE.equals(includeTestPatients)) {
            this.appendWhereClause(whereClause, SqlUtil.generateExcludeTestPatientsClause(true, "a", "patientSsn"));
        }
		
		if (!NullChecker.isNullOrEmpty(sortValue)
				&& !NullChecker.isNullOrEmpty(sortDirection)) {
			if (this.validColumnHeadersAuditedConsent(sortValue)
					&& this.validSortDirection(sortDirection)) {
				if ("timeOfEvent".equals(sortValue) || 
						"createdDate".equals(sortValue)) {
					sortClause.append("a." + sortValue + " " + sortDirection.toUpperCase());
				}
                                else if ("visnName".equals(sortValue)) {
                                        sortClause.append("NLSSORT(a." + sortValue + ", 'NLS_SORT=LATIN') " + sortDirection.toUpperCase());
                                }
				else  {
					sortClause
					.append("upper(a." + sortValue + ") " + sortDirection.toUpperCase())
					.append(", a.timeOfEvent DESC");
				}
			}
		} else {
			sortClause.append("a.timeOfEvent DESC"); //default sort
		}
                
		final Query query = this.makeAuditedConsentQuery(whereClause,
				sortClause);
		// + " ORDER BY a.timeOfEvent DESC");

		for (final Map.Entry<String, Object> entry : setParams.entrySet()) {
			String pKey = entry.getKey();
			Object pValue = entry.getValue();
			if (pKey.equalsIgnoreCase("startDate") || pKey.equalsIgnoreCase("endDate")) {
				query.setParameter(pKey, (Date)pValue, TemporalType.TIMESTAMP);
			} else {
				query.setParameter(pKey, pValue);
			}
		}

		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		return this.getResults(updatedQuery, currentPage, pageSize);
	}

	/**
	 * Get the events from the database table by calling ConsentAudit.findByAll
	 * query.
	 */
        public final List<Object> getEvents(final int currentPage,
			final int pageSize) {

		final String reportQueryName = "ConsentAudit.findAll";
		final Query query = this.em.createNamedQuery(reportQueryName);

		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		return this.getResults(updatedQuery, currentPage, pageSize);
	}

	/**
	 * Get the events from the database table by calling
	 * ConsentAudit.findByPatientId query.
	 */
        public final List<Object> getEvents(final String patientId,
			final int currentPage, final int pageSize) {

		final String reportQueryName = "ConsentAudit.findByPatientId";
		final Query query = this.em.createNamedQuery(reportQueryName)
				.setParameter("patientId", patientId);

		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		return this.getResults(updatedQuery, currentPage, pageSize);
	}

	/**
	 * Get the event summaries from the database table by building and calling a
	 * query. Use this method for complex queries resulting in a summary. It
	 * accepts all parameters and builds the where and order by clauses by the
	 * values supplied.
	 */
	public final List<AuditedConsentSummary> getEventsSummary(
			final Date startDate, final Date endDate, final String facilityName,
			final String consentType, final int currentPage, final int pageSize,

			final String sortBy, final String sortDirection, final Boolean includeTestPatients) {


		final StringBuffer whereClause = new StringBuffer();
		final StringBuffer groupByClause = new StringBuffer();
		final HashMap<String, Object> setParams = new HashMap<String, Object>();

		if (startDate != null) {
			this.appendWhereClause(whereClause, "a.timeOfEvent >= :startDate");
			setParams.put("startDate", startDate);
		}

		if (endDate != null) {
			this.appendWhereClause(whereClause, "a.timeOfEvent <= :endDate");
			setParams.put("endDate", endDate);
		}

		if (!NullChecker.isNullOrEmpty(facilityName)) {
			if (facilityName.equalsIgnoreCase("Unknown")) {
				this.appendWhereClause(whereClause, "a.facilityName IS NULL");
			} else {
				this.appendWhereClause(whereClause, "a.facilityName = :facilityName");
				setParams.put("facilityName", facilityName);
			}
		}
        

        if (!Boolean.TRUE.equals(includeTestPatients)) {

            this.appendWhereClause(whereClause, 
                    "((a.patientSsn NOT LIKE '000%' AND a.patientSsn NOT LIKE '666%') OR " +
                    "a.patientSsn IS NULL)");
        }

		if (!NullChecker.isNullOrEmpty(consentType)) {
			if (consentType.equalsIgnoreCase("allAuthorizations")){
				this.appendWhereClause(whereClause, "a.actionType = :actionType");
				setParams.put("actionType", "OPT-IN");
			} else if(consentType.equalsIgnoreCase("allRevocations")){
				this.appendWhereClause(whereClause, "a.actionType = :actionType");
				setParams.put("actionType", "OPT-OUT");
			} else if (consentType.equalsIgnoreCase("NwHINAuthorizations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Authorization");
			} else if (consentType.equalsIgnoreCase("SSAAuthorizations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "SSA Authorization");
			} else if (consentType.equalsIgnoreCase("NwHINRevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Revocation");
			} else if (consentType.equalsIgnoreCase("SSARevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "SSA Revocation");
			} else if (consentType.equalsIgnoreCase("NwHINRestrictions")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Organization Restriction Authorization");
			} else if (consentType.equalsIgnoreCase("NwHINRestrictionRevocations")){
				this.appendWhereClause(whereClause, "a.consentType = :consentType");
				setParams.put("consentType", "NwHIN Organization Restriction Revocation");
			} else {
				//TODO throw an error
				//program did not receive expected value
			}
		}

		this.appendGroupByClause(groupByClause, "a.facilityName, a.consentType");

		final String filteredSortBy;
		final Query query;
		if (!NullChecker.isNullOrEmpty(sortBy)
				&& !NullChecker.isNullOrEmpty(sortDirection)
				&& this.validColumnHeadersAuditedConsentSummary(sortBy)
				&& this.validSortDirection(sortDirection)) {
			if (sortBy.equalsIgnoreCase("consentType")) {
				filteredSortBy = "a.consentType";
			} else if (sortBy.equalsIgnoreCase("facilityName")) {
				filteredSortBy = "a.facilityName";
			} else {
				filteredSortBy = "COUNT(*)";
			}
			final String updatedQuery = new String(
					"SELECT NEW gov.va.nvap.web.consent.audit.AuditedConsentSummary "
							+ "(a.consentType, a.facilityName, COUNT(a.consentType)) "
							+ "FROM AuditedConsent a " + whereClause.toString()
							+ groupByClause.toString() + " ORDER BY "
							+ filteredSortBy.toString() + " "
							+ sortDirection.toUpperCase().toString());
			// TODO add additional orderby clause on created date to all results
			// exported to excel look the same on the screen
			query = this.em.createQuery(updatedQuery);
		} else {
			query = this.em
					.createQuery("SELECT NEW gov.va.nvap.web.consent.audit.AuditedConsentSummary (a.consentType, a.facilityName, COUNT(a.consentType)) "
							+ "FROM AuditedConsent a "
							+ whereClause.toString()
							+ groupByClause.toString()
							+ " ORDER BY a.facilityName ASC");
		}

		for (final Map.Entry<String, Object> entry : setParams.entrySet()) {
			String pKey = entry.getKey();
			Object pValue = entry.getValue();
			if (pKey.equalsIgnoreCase("startDate") || pKey.equalsIgnoreCase("endDate")) {
				query.setParameter(pKey, (Date)pValue, TemporalType.TIMESTAMP);
			} else {
				query.setParameter(pKey, pValue);
			}
		}

		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
                
		Assert.assertNotEmpty(updatedQuery, "Query cannot be empty");
		List<AuditedConsentSummary> summaryList;
		if (pageSize > 0) {
			summaryList = updatedQuery.setFirstResult(currentPage * pageSize)
					.setMaxResults(pageSize).getResultList();
		} else {
			summaryList = updatedQuery.getResultList();
		}

/*		final List<AuditedConsentSummary> summaryListOptInOut = new ArrayList<AuditedConsentSummary>();
		// filter out non opt-in opt-out action type
		for (final AuditedConsentSummary auditedConsentSummary : summaryList) {
			if (auditedConsentSummary.getConsentType().equals(
					ConsentAuditType.OPTIN_AUDIT_TYPE.value())
					|| auditedConsentSummary.getConsentType().equals(
							ConsentAuditType.OPTOUT_AUDIT_TYPE.value())) {
				summaryListOptInOut.add(auditedConsentSummary);
			}
		}

		return summaryListOptInOut;*/
		return summaryList;
	}

	/**
	 * Get the message content from the table. We do not want hibernate to pull
	 * all the records with the clob content. So have an independent query to
	 * just pull messages from the clob.
	 */
	public final String getMessageContent(final Long consentAuditId) {
		Assert.assertNotEmpty(consentAuditId,
				"Audited event id cannot be empty");
		// Query to get the individual clob content
		final Query query = this.em
				.createQuery("SELECT ca.detail from ConsentAudit ca where ca.consentAuditId = :consentAuditId");
		query.setParameter("consentAuditId", consentAuditId);

		final Query updatedQuery = this.setAdditionalQueryParamaters(query);
		//if(Clob.class.isInstance(Clob.class)) {
			//final Clob clob = (Clob) updatedQuery.getSingleResult();
		try {
			final String mydetail = (String) updatedQuery.getSingleResult();	
			if (NullChecker.isNotEmpty(consentAuditId) && !NullChecker.isNullOrEmpty(mydetail)) {
				final StringReader reader = new StringReader(mydetail);
				if (NullChecker.isEmpty(reader)) {
					return null;
				}
				final StringBuffer sb = new StringBuffer();
				final char[] charbuf = new char[AuditedConsentDAO.READ_BUFFER_SIZE];
				for (int i = reader.read(charbuf); i > 0; i = reader
						.read(charbuf)) {
					sb.append(charbuf, 0, i);
				}
				// Return the CDA R2 XML
				return sb.toString();
			}
		} catch (final IOException e) {
			throw new RuntimeException(e);
		}
		//}
		return null;
	}
        
	/**
	 * Get the results from the database table based on the patient id.
	 */
        public List<Object> getResultsByPatientId(String patientId) {
            Query query = this.em.createQuery("SELECT a FROM AuditedConsentEx a WHERE a.patientId = :patientId ORDER BY a.createdDate DESC");
            query.setParameter("patientId", patientId);

            return query.getResultList();
        }
   
	/**
	 * Get the results from the database table.
	 */
        public final List<Object> getResults(final Query query,
			final int currentPage, final int pageSize) {
		Assert.assertNotEmpty(query, "Query cannot be empty");
                List<Object> auditedEventEntityList;
		if (pageSize > 0) {
			auditedEventEntityList = query
					.setFirstResult(currentPage * pageSize)
					.setMaxResults(pageSize).getResultList();
		} else {
			auditedEventEntityList = query.getResultList();
		}
		return auditedEventEntityList;
	}

	private Query makeAuditedConsentQuery(final StringBuffer whereClause,
			final StringBuffer sortClause) {
		if (NullChecker.isNullOrEmpty(sortClause)) {
                        return this.em.createQuery("SELECT a FROM AuditedConsentEx a "
                                        + whereClause.toString());
		} else {
			return this.em.createQuery("SELECT a FROM AuditedConsentEx a "
                                        + whereClause.toString() + " ORDER BY "
					+ sortClause.toString());
		}
	}

	/**
	 * Optimize the query parameters for fetching the result.
	 */
	private Query setAdditionalQueryParamaters(final Query theQuery) {
		theQuery.setHint("org.hibernate.cacheable", Boolean.TRUE)
				.setHint("org.hibernate.fetchSize", Integer.valueOf(50))
				.setHint("org.hibernate.cacheMode", CacheMode.GET)
				.setHint("org.hibernate.readOnly", Boolean.TRUE)
				.setHint("org.hibernate.flushMode", FlushMode.MANUAL);
		return theQuery;
	}

	/**
	 * Set the entity manager (Injected by Spring).
	 */
	public void setEntityManager(final EntityManager em) {
		this.em = em;
	}

	private boolean validColumnHeadersAuditedConsent(final String header) {
		if (header.equalsIgnoreCase("consentAuditId")
				|| header.equalsIgnoreCase("actionType")
				|| header.equalsIgnoreCase("consentType")
				|| header.equalsIgnoreCase("userId")
				|| header.equalsIgnoreCase("facility")
				|| header.equalsIgnoreCase("facilityName")
				|| header.equalsIgnoreCase("visnName")
				|| header.equalsIgnoreCase("patientId")
				|| header.equalsIgnoreCase("patientLastName")
				|| header.equalsIgnoreCase("patientGivenName")
				|| header.equalsIgnoreCase("patientMiddleName")
				|| header.equalsIgnoreCase("patientSsn")
				|| header.equalsIgnoreCase("timeOfEvent")
				|| header.equalsIgnoreCase("createdDate")
				|| header.equalsIgnoreCase("optoutReason")
				|| header.equalsIgnoreCase("pouValue")) {
			return true;
		}
		throw new InvalidParameterException("Wrong column header to sort on: "
				+ header);
	}

	/**
	 * Checks validity of columns to sort on for Summary report
	 */
	private boolean validColumnHeadersAuditedConsentSummary(final String sortBy) {
		if (sortBy.equalsIgnoreCase("consentType")
				|| sortBy.equalsIgnoreCase("facilityName")
				|| sortBy.equalsIgnoreCase("count_All_Summary_Report")) {
			return true;
		} else {
			throw new InvalidParameterException(
					"Wrong column header to sort on in summary report: "
							+ sortBy);
		}
	}

	private boolean validSortDirection(final String sortDirection) {
		if (sortDirection.equalsIgnoreCase("ASC")
				|| sortDirection.equalsIgnoreCase("DESC")) {
			return true;
		}
		return false;
	}
}
