package gov.va.nvap.web.helper.report;

import gov.va.nvap.common.validation.NullChecker;
import gov.va.nvap.service.pdq.PatientMatchQuery;
import gov.va.nvap.service.pdq.PatientMatchResponse;
import gov.va.nvap.service.pdq.PatientProfile;
import gov.va.nvap.service.pdq.PdqException;
import gov.va.nvap.service.pdq.PdqService;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

import org.springframework.beans.factory.annotation.Required;

public class ReportHelper {

	private PdqService pdqService;
	private String disclosuresDetailedReportCap;
	private String recvdDocsDetailedReportCap;
	private String consentDirectivesReportCap;
	private String pdDetailedReportCap;
		
	public String resolveICN(final String firstName, final String lastName,
			final String ssn) {

		if (NullChecker.isNotEmpty(ssn)
				&& NullChecker.isNotEmpty(firstName)
				&& NullChecker.isNotEmpty(lastName)) {

			final List<PatientProfile> patients = getMatchedPatients(ssn, firstName, lastName);
			if(patients != null) {
				if (patients.isEmpty()) {
					return "Unknown";
				} else if (patients.size() >= 2) {
					return "Multiple ICNs";
				} else {
					final String icn = patients.get(0).getIcn();
					return icn;
				}
			}
		} else {
			return "Unknown";
		}

		return "Unknown";
	}

	/**
	 * This method returns list of Matched Patients.
	 * 
	 * @param ssn
	 * @param firstName
	 * @param lastName
	 * @return List of Matched Patients
	 */
	public List<PatientProfile> getMatchedPatients(String ssn, String firstName, String lastName) {

		List<PatientProfile> patients = null;
		try {

			// Create a patient profile and search for patient
			final PatientProfile patient = new PatientProfile();
			patient.setFirstName(firstName);
			patient.setLastName(lastName);
			patient.setSsn(ssn);
			final PatientMatchQuery patientSearchRequest = new PatientMatchQuery();
			patientSearchRequest.setPatientProfile(patient);
			// Call PDQ service to get the matching patient
			final PatientMatchResponse patientSearchResponse = this.pdqService
					.searchPatient(patientSearchRequest);
			patients = patientSearchResponse
					.getPatients().getPatientProfile();

		} catch (final PdqException ex) {
			// TODO: Maybe enter a message that MPI is not available to resolve
			// the ICN??
			// throw new ServletException(ex);
			return patients;
		}

		return patients;
	}
		
	/**
	 * Checks whether the given SSN is for a test patient.
	 * 
	 * @param ssn SSN to check
	 * @return true or false
	 */
	public boolean isTestPatient(String ssn) {
        return NullChecker.isNotEmpty(ssn) &&        // Ignore null entries
                ssn.length() == 9 && (               // Ignore entries that aren't exactly 9 digits
                ssn.startsWith("000") ||             // Area number 000 denotes a test patient
                ssn.startsWith("666") ||             // Area number 666 denotes a test patient
                ssn.startsWith("9") ||               // Area number in 900-999 range denotes a test patient
                ssn.substring(3, 5).equals("00") ||  // Group number 00 denotes a test patient
                ssn.substring(5, 9).equals("0000")); // Serial number 0000 denotes a test patient
	}
	
	public String getFormattedDateTime(Date date) {
		return getFormattedDateString(date,
				"MM/dd/yyyy hh:mm a");
	}
	
	public String getFormattedDate(Date date) {
		return getFormattedDateString(date,
				"MM/dd/yyyy");
	}
	
	public int getDisclosuresDetailedReportCapAsInt() {
		return Integer.parseInt(this.disclosuresDetailedReportCap);
	}

	public int getRecvdDocsDetailedReportCapAsInt() {
		return Integer.parseInt(this.recvdDocsDetailedReportCap);
	}

	public int getConsentDirectivesReportCapAsInt() {
		return Integer.parseInt(this.consentDirectivesReportCap);
	}

	public int getPdDetailedReportCapAsInt() {
		return Integer.parseInt(this.pdDetailedReportCap);
	}

	@Required
	public void setPdqService(final PdqService pdqService) {
		this.pdqService = pdqService;
	}

	@Required
	public void setDisclosuresDetailedReportCap(final String disclosuresDetailedReportCap) {
		this.disclosuresDetailedReportCap = disclosuresDetailedReportCap;
	}

	@Required
	public void setRecvdDocsDetailedReportCap(final String recvdDocsDetailedReportCap) {
		this.recvdDocsDetailedReportCap = recvdDocsDetailedReportCap;
	}

	@Required
	public void setConsentDirectivesReportCap(final String consentDirectivesReportCap) {
		this.consentDirectivesReportCap = consentDirectivesReportCap;
	}

	@Required
	public void setPdDetailedReportCap(final String pdDetailedReportCap) {
		this.pdDetailedReportCap = pdDetailedReportCap;
	}
	
	private String getFormattedDateString(final Date date, final String datePattern) {
		if (date == null) {
			return "Unknown";
		}

		final SimpleDateFormat formatter = new SimpleDateFormat(
				datePattern);
		
		return formatter.format(date);

	}
	
	  /**
     * Extracts value of key=value pair from details property
     * @param details Details property of AuditType
     * @param key Key of the key/value pair
     * @return Value of the key/value pair
     */
	public String extractFromDetails(String details, String key) {
        int pos1 = details.indexOf(key + "=");
        int pos2;
        
        if (pos1 != -1) {
            pos2 = details.indexOf(",", pos1);
            if (pos2 != -1) {
                return details.substring(pos1 + key.length() + 1, pos2);
            }
        }
        return "";
    }

}
