package gov.va.nvap.web.report;

import gov.va.med.nhin.adapter.audit.FieldType;
import gov.va.med.nhin.adapter.audit.PageInfoType;
import gov.va.med.nhin.adapter.audit.SortDirection;
import gov.va.med.nhin.adapter.audit.SortFieldType;
import gov.va.med.nhin.adapter.audit.SortFieldsType;
import gov.va.nvap.common.validation.NullChecker;
import gov.va.nvap.privacy.ConsentDirectiveDetailedExpirationRequest;
import gov.va.nvap.privacy.ConsentDirectiveDetailedExpirationResponse;
import gov.va.nvap.privacy.ConsentDirectiveReferenceType;
import gov.va.nvap.privacy.DetailedConsentDirectiveReferenceType;
import gov.va.nvap.service.adapter.audit.AdapterAuditManager;
import gov.va.nvap.service.adapter.doc.AdapterException;
import gov.va.nvap.service.pdq.PatientDemographics;
import gov.va.nvap.web.app.ResponseDispatcherHttpServlet;
import gov.va.nvap.web.helper.document.DocumentHelper;
import gov.va.nvap.web.helper.privacy.ConsentManagementHelper;
import gov.va.nvap.web.helper.report.ReportHelper;
import gov.va.nvap.web.util.Constants;
import gov.va.nvap.web.util.Paginator;
import gov.va.nvap.web.util.xls.ExcelExporter;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ejb.EJB;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.poi.ss.usermodel.Workbook; 

/**
 * @author Zack Peterson
 */
public class ExpiringConsentReportResults extends ResponseDispatcherHttpServlet {
	
	/**
	 * Serial UID.
	 */
	private static final long serialVersionUID = -4614454894945125740L;
    
    private AdapterAuditManager adapterAuditManager;
	
    /**
	 * Get the Consent Management Service Helper class from Spring. The
     * Helper class can be used to access consent management functions on
     * the server side.
	 * 
	 * @return helper class for consent management purposes
	 */
	public ConsentManagementHelper getCmsHelper() {
		final ConsentManagementHelper cmsHelper = this.getBean("cmsHelper",
				ConsentManagementHelper.class);
		return cmsHelper;
	}

        public void exportToExcel(final HttpServletRequest request, 
             final HttpServletResponse response) throws ServletException, IOException {
            
            final HttpSession session = request.getSession(false);
            final List<Map<String, Object>> results = this.getResults(
            request.getSession(), new Paginator(this.getReportHelper().getConsentDirectivesReportCapAsInt()));
            
            // Create a map of key which is based on the result data key and the heading value 
            //The heading is used to create the column headers and the
            //key is used to pull the data from the results 
            final Map<String,String> reportMap = new LinkedHashMap<String, String>();
            
            // Set the heading parameters 
            final Map<String, List<Object>> headingParameters = new LinkedHashMap<String, List<Object>>();
            
            String header;
            
            for (Map<String, Object> entry: results){
                SimpleDateFormat dt = new SimpleDateFormat("MM/dd/yyyy");
                Date oldOptInDate = (Date)entry.get("optInDate");
                String newOptInDate = dt.format(oldOptInDate);
                entry.put("optInDate", newOptInDate);
                
                Date oldExpirationDate = (Date)entry.get("expirationDate");
                String newExpirationDate = dt.format(oldExpirationDate);
                entry.put("expirationDate", newExpirationDate);
            }
            
            //Check to see if this is an OptIn Report. The session attribute gets 
            //set in optInReportResults.jsp. 
            boolean optInReport;
            try{
                if(session.getAttribute("optInReport").equals("true")){
                    optInReport = true;
                    header = "Opt-In Patients Detailed Report";
                }
                else{
                    optInReport = false;
                    header = "Expiring Consent Detailed Report";
                }
            }
            catch(NullPointerException e){
                optInReport = false;
                header = "Expiring Consent Detailed Report";
            }
            
            List<Object> val;
            
            //Check if this this is an Expiring Consent report in which case the 
            //start date and end dates MUST be present. Otherwise this is an Opt In
            //report and dates CANNOT be provided.
            if (!optInReport){
                final Object startDateStr = session.getAttribute("startDate");
                val = new ArrayList<Object>();
                val.add(this.getReportHelper().getFormattedDate((Date)startDateStr));
                headingParameters.put("StartDate:", val);
                
                final Object endDateStr = session.getAttribute("endDate");
                val = new ArrayList<Object>();
                val.add(this.getReportHelper().getFormattedDate((Date)endDateStr));
                headingParameters.put("EndDate:", val);
            }
            
            LinkedHashMap<String,Object> filters = new LinkedHashMap<String,Object>();
            //List all the possible optional filters for Opt In Report and
            //Expiring Consent Report.
            filters.put("Entered By",session.getAttribute("userId"));
            filters.put("VA Facility",session.getAttribute("facilityName"));
           
            String consentType = (String)session.getAttribute("consentTypeName");
            if (consentType.contains("NwHIN")){
                consentType = "eHealth Exchange Authorization";
            }
            filters.put("Consent Type",consentType);
            
            //Iterate through the possible filters and remove empty filters.
            for (Map.Entry<String,Object> filter : filters.entrySet()){
                String filterName = filter.getKey();
                Object value = filter.getValue();
                List<Object> headingValue = new ArrayList<Object>();
                if (!(value == null) && !("".equals(value)) && !("Unknown".equals(value))){
                    headingValue.add(value);
                    headingParameters.put(filterName, headingValue);
                }
            }    
            
            if ((Boolean)session.getAttribute("testPatients")){
                List<Object> headingValue = new ArrayList<Object>();
                headingValue.add("Yes");
                headingParameters.put("Include test patients", headingValue);
            }
            
            if (NullChecker.isNullOrEmpty(results)){
                reportMap.put("", "No records were found.");
            }
            
            else {
                reportMap.put("ssnMasked", "SSN"); 
                reportMap.put("lastName", "Patient Last Name"); 
                reportMap.put("firstName", "Patient First Name");
                reportMap.put("middleName", "Patient Middle Name");
                reportMap.put("optInDate", "Opt In Date");
                reportMap.put("expirationDate", "Expiration Date");
                reportMap.put("consentTypeName", "Consent Type");
                reportMap.put("userId", "Entered By"); 
                reportMap.put("facilityName","VA Facility");
            }
            
            // Create workbook 
            final Workbook wb = this.getExcelExporter().exportToExcel(
            header, header,
            reportMap, results, headingParameters, null);
            
            // Write Excel to Stream
            this.getExcelExporter().writeExcelToStream(wb, response); 
            
        }
        
	public DocumentHelper getDocumentHelper() {
		return this.getBean("adapterDocumentHelper", DocumentHelper.class);
	}

	/**
	 * Get the excel exporter class from Spring.
	 * 
	 * @return the excel exporter object
	 */
	public ExcelExporter getExcelExporter() {
		final ExcelExporter excelExporter = this.getBean("excelExporter",
				ExcelExporter.class);
		return excelExporter;
	}

	public ReportHelper getReportHelper() {
		final ReportHelper reportHelper = this.getBean("reportHelper",
				ReportHelper.class);
		return reportHelper;
	}

    /**
	 * Gathers and assembles the information for the Expiring Consent Report.
	 * 
     * @param session   the http session
     * @param paginator the paginator for the expiring consent report
	 * @return a list of map objects with patient information for the report
	 */
	private List<Map<String, Object>> getResults(HttpSession session,
			final Paginator paginator) {
        
        Date startDate = (Date) session.getAttribute("startDate");
        Date endDate = (Date) session.getAttribute("endDate");
		final ConsentDirectiveDetailedExpirationRequest cdExpirationReq = new ConsentDirectiveDetailedExpirationRequest();
		
		cdExpirationReq.setStartDate(startDate);
        cdExpirationReq.setEndDate(endDate);
        cdExpirationReq.setFacilityName((String) session.getAttribute("facilityName"));
        cdExpirationReq.setConsentTypeName((String) session.getAttribute("consentTypeName"));
        cdExpirationReq.setUserId((String) session.getAttribute("userId"));
        
		final String inboundSortValue = (String) session
				.getAttribute("inboundSortValue");
		final String inboundSortDirection = (String) session
				.getAttribute("inboundSortDirection");
        
        final Boolean includeTestPatients = (Boolean) session
				.getAttribute("testPatients");
        
        cdExpirationReq.setTestPatients(includeTestPatients);

		// For use in sorting results
		if (NullChecker.isNotEmpty(inboundSortValue)) {
			final SortFieldType sortField = new SortFieldType();
			sortField.setField(FieldType.fromValue(inboundSortValue));
			if ("DESC".equalsIgnoreCase(inboundSortDirection)) {
				sortField.setDirection(SortDirection.DESC);
			} else {
				sortField.setDirection(SortDirection.ASC);
			}
			final SortFieldsType sortFieldsType = new SortFieldsType();
			sortFieldsType.getSortField().add(sortField);
			cdExpirationReq.setSortFields(sortFieldsType);
		}

		// Set Page Information
		final PageInfoType pageInfoType = new PageInfoType();
		pageInfoType.setPageNumber(paginator.getCurrentPage());
		pageInfoType.setPageSize(paginator.getRecordsPerPage());
		cdExpirationReq.setPageInfo(pageInfoType);

		final List<Map<String, Object>> results = new ArrayList<Map<String, Object>>();
		try {

            ConsentDirectiveDetailedExpirationResponse detailedCmsResponse = this
					.getCmsHelper().getAllExpiringDetailedConsentDirectives(cdExpirationReq);
                        
			if (detailedCmsResponse != null) {
                
                List<DetailedConsentDirectiveReferenceType> detailedConsentReferences =
                        detailedCmsResponse.getDetailedConsentDirectiveReference();

				if ((detailedConsentReferences != null)
						&& (detailedConsentReferences.size() > 0)) {
					
					Long totalRows = detailedCmsResponse.getTotalRows();
					session.setAttribute("totalRows", totalRows);
					session.setAttribute("expiredNotificationLettersMaxTime", Constants.getExpiredNotificationLettersMaxTime());
					session.setAttribute("expiredNotificationPerLetterTime", Constants.getExpiredNotificationPerLetterTime());
					int pageNumber = paginator.getCurrentPage();
					int pageSize = paginator.getRecordsPerPage();
					int toPage = 0;
					boolean nextDisabled = false;
					
					if(pageNumber <= 0) {
						toPage = pageSize;
					} else {
						toPage = (pageNumber+1) * pageSize;
					}
					if(null != totalRows) {
						if(toPage == totalRows.intValue()) {
							nextDisabled = true;
						}
					}					
					paginator.setNextDisabled(nextDisabled);
					session.setAttribute("nextDisabled", nextDisabled);
					
                    for (DetailedConsentDirectiveReferenceType detailedConsentReference : detailedConsentReferences) {
                        final Map<String, Object> resultMap = new HashMap<String, Object>();
                        if (detailedConsentReference != null ) {
                            ReportDataProcessor.addSsnToResultMap(resultMap, detailedConsentReference.getSsn());
                            resultMap.put("lastName", detailedConsentReference.getLastName());
                            resultMap.put("firstName", detailedConsentReference.getFirstName());
                            resultMap.put("middleName", detailedConsentReference.getMiddleName());
                            String consentTypeName = detailedConsentReference.getOptinConsentType().value();
                            if (consentTypeName.contains("NwHIN")) {
                                consentTypeName = consentTypeName.replace("NwHIN", Constants.getOrganizationName());
                            }
                            resultMap.put("optInDate", detailedConsentReference.getOptinDate());
                            resultMap.put("expirationDate", detailedConsentReference.getExpirationDate());
                            resultMap.put("facilityName", detailedConsentReference.getFacilityName());
                            resultMap.put("userId", detailedConsentReference.getUserId());
                            results.add(resultMap);
                        }
                    }
				}
			}
		} catch (final Exception ex) {
			throw new RuntimeException(ex);
		}
        
		return results;
	}

	private boolean isPaginatorPresent(HttpSession session) {
		return !NullChecker.isNullOrEmpty(session.getAttribute("paginator"));
	}

	public void next(HttpServletRequest request,
			final HttpServletResponse response) throws ServletException,
			IOException {
		HttpSession session = request.getSession(false);
		if (this.isPaginatorPresent(session)) {
			Paginator paginator = (Paginator) session
					.getAttribute("paginator");
			paginator.next(request);
			session.setAttribute("paginator", paginator);
			session.setAttribute("expiringConsentReportResults",
					this.getResults(session, paginator));
			
			this.forward(request, response, "show");
		} else {
			this.forward(request, response, "noshow");
		}
	}

	public void prev(HttpServletRequest request,
			final HttpServletResponse response) throws ServletException,
			IOException {
		HttpSession session = request.getSession(false);
		if (this.isPaginatorPresent(session)) {
			Paginator paginator = (Paginator) session
					.getAttribute("paginator");
			paginator.previous(request);
			session.setAttribute("paginator", paginator);
			session.setAttribute("expiringConsentReportResults",
					this.getResults(session, paginator));
			this.forward(request, response, "show");
		} else {
			this.forward(request, response, "noshow");
		}
	}

	@EJB(beanInterface = AdapterAuditManager.class, mappedName = "AdapterAuditManager")
	public void setAdapterAuditManager(
			final AdapterAuditManager adapterAuditManager) {
		this.adapterAuditManager = adapterAuditManager;
	}

	@Override
	protected void unspecified(final HttpServletRequest request,
			final HttpServletResponse response) throws ServletException,
			IOException {
		final HttpSession session = request.getSession(false);

		// sort
		final String expiringConsentSortValue = request
				.getParameter("inboundSortValue");
		final String expiringConsentSortDirection = request
				.getParameter("inboundSortDirection");
		session.setAttribute("inboundSortValue", expiringConsentSortValue);
		session.setAttribute("inboundSortDirection",
				expiringConsentSortDirection);

		if (this.isPaginatorPresent(session)) {
			Paginator paginator = (Paginator) session.getAttribute("paginator");
			if (NullChecker.isEmpty(paginator)) {
				paginator = new Paginator(0);
				session.setAttribute("paginator", paginator);
			}
			paginator.reset();
			session.setAttribute("expiringConsentReportResults",
					this.getResults(session, paginator));
			this.forward(request, response, "show");
		} else {
			this.forward(request, response, "noshow");
		}
	}

	public void view(HttpServletRequest request,
			final HttpServletResponse response) throws ServletException,
			IOException {
		HttpSession session = request.getSession(false);
		final String icn = request.getParameter("icn");
		final String documentUniqueId = request
				.getParameter("expiringConsentUniqueId");
		final String styleSheetViewType = request
				.getParameter("styleSheetViewType");

		String document = null;
		try {
			document = this.getDocumentHelper().getDocument(icn,
					documentUniqueId, request.getRemoteUser(),
					this.getCmsHelper().getCompleteHomeCommunityId());
		} catch (final AdapterException ex) {
			throw new ServletException(ex);
		}

		session.setAttribute("icn", icn);
		session.setAttribute("documentUniqueId", documentUniqueId);
		session.setAttribute("document", document);
		session.setAttribute("isXmlViewEnabled", this.getDocumentHelper()
				.isXmlViewEnabled());

		if ("xml".equals(styleSheetViewType)) {
			document = document
					.replaceAll(
							"<?xml-stylesheet type=\"text/xsl\" href=\"CCD.xsl\"?>",
							"");
			response.setContentType("text/xml;charset=UTF-8");
			response.getWriter().write(document);
		} else {
			// Update session in the document
			final String html = this.getDocumentHelper().getHtml(document,
					styleSheetViewType);

			session.setAttribute("updatedDocument", html);
			this.forward(request, response, "showInboundDocumentReportDocument");
		}
	}
    
    public void download(final HttpServletRequest request,
			final HttpServletResponse response) throws ServletException,
			IOException {
        
        final HttpSession session = request.getSession(false);
        Date startDate = (Date) session.getAttribute("startDate");
        Date endDate = (Date) session.getAttribute("endDate");
        
        final ByteArrayInputStream byteInputStream = 
                this.getCmsHelper().getExpiringConsentReportLetters(startDate, endDate);
        
        response.setContentType("application/pdf");
        response.addHeader("Content-Disposition", "attachment; filename=patient_letters.pdf");
        
        OutputStream responseOutStream = response.getOutputStream();
        
        byte[] buffer = new byte[4096];
        int bytesRead;
        int totalBytes = 0;
        
        while ((bytesRead = byteInputStream.read(buffer)) != -1) {
            responseOutStream.write(buffer, 0, bytesRead);
            totalBytes += bytesRead;
        }
        
        response.setContentLength(totalBytes);
        
        byteInputStream.close();
        responseOutStream.flush();
        
        this.forward(request, response, "success");
    }
}
